from functools import wraps
from typing import Dict
from typing import Optional

from flask import request
from flask import session


class Session(object):
    _current_language_key = 'current_language'
    _last_visited_url_key = 'last_visited_url'

    @classmethod
    def _session(cls) -> Dict[str, str]:
        return session

    @classmethod
    def store_last_visited_url(cls):
        cls._session()[cls._last_visited_url_key] = request.url

    @classmethod
    def get_last_visited_url(cls) -> Optional[str]:
        return cls._session().get(cls._last_visited_url_key)

    @classmethod
    def store_current_language(cls, language: str):
        cls._session()[cls._current_language_key] = language

    @classmethod
    def get_current_language(cls) -> str:
        return cls._session().get(cls._current_language_key)


def track_history(func):
    @wraps(func)
    def history_tracker(*args, **kwargs):
        Session.store_last_visited_url()
        return func(*args, **kwargs)

    return history_tracker



from datetime import datetime
from io import BytesIO
from itertools import chain
from mimetypes import guess_type
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from flask import render_template
from flask import request
from flask_login import current_user
from flask_wtf import FlaskForm
from PIL import Image
from werkzeug.datastructures import FileStorage
from wtforms import FileField
from wtforms import SelectMultipleField
from wtforms import StringField
from wtforms import SubmitField
from wtforms.validators import DataRequired
from wtforms.validators import Optional as DataOptional

from opwen_email_client.domain.email.store import EmailStore
from opwen_email_client.util.wtforms import Emails
from opwen_email_client.util.wtforms import HtmlTextAreaField
from opwen_email_client.webapp.config import AppConfig
from opwen_email_client.webapp.config import ImageDimensions
from opwen_email_client.webapp.config import i8n


class NewEmailForm(FlaskForm):
    to = StringField(validators=[
        DataRequired(i8n.EMAIL_TO_REQUIRED),
        Emails(AppConfig.EMAIL_ADDRESS_DELIMITER, i8n.EMAIL_ADDRESS_INVALID),
    ])

    cc = StringField(validators=[
        DataOptional(),
        Emails(AppConfig.EMAIL_ADDRESS_DELIMITER, i8n.EMAIL_ADDRESS_INVALID),
    ])

    bcc = StringField(validators=[
        DataOptional(),
        Emails(AppConfig.EMAIL_ADDRESS_DELIMITER, i8n.EMAIL_ADDRESS_INVALID),
    ])

    subject = StringField(validators=[DataOptional()])

    body = HtmlTextAreaField(validators=[
        DataOptional(),
    ])

    forwarded_attachments = SelectMultipleField(choices=[], validators=[
        DataOptional(),
    ])

    attachments = FileField(validators=[
        DataOptional(),
    ], render_kw={
        'multiple': True,
    })

    submit = SubmitField()

    def as_dict(self, email_store: EmailStore) -> dict:
        form = {key: value for (key, value) in self.data.items() if value}
        form.pop('submit', None)

        attachments = request.files.getlist(self.attachments.name)
        to = _split_emails(form.get('to'))
        cc = _split_emails(form.get('cc'))
        bcc = _split_emails(form.get('bcc'))

        sent_at = None
        if all(_is_local_message(address) for address in chain(to, cc, bcc)):
            sent_at = datetime.utcnow().strftime('%Y-%m-%d %H:%M')

        form['sent_at'] = sent_at
        form['read'] = True
        form['from'] = current_user.email
        form['to'] = to
        form['cc'] = cc
        form['bcc'] = bcc
        form['body'] = form.get('body')
        form['subject'] = form.get('subject', i8n.EMAIL_NO_SUBJECT)
        form['attachments'] = list(_attachments_as_dict(attachments))
        return form

    def _populate(self, email_store: EmailStore):
        pass

    @classmethod
    def _new_instance_for(cls, action_name: Optional[str]):
        if not action_name:
            return cls(request.form)

        try:
            clazz = next(clazz for clazz in NewEmailForm.__subclasses__()
                         if getattr(clazz, 'action_name', None) == action_name)
        except StopIteration:
            return None
        else:
            return clazz(request.form)

    def _get_reference_email(self, email_store: EmailStore) -> Optional[dict]:
        uid = request.args.get('uid', '')
        if not uid:
            return None

        reference = email_store.get(uid)
        if not self._can_access(current_user, reference):
            return None

        return reference

    @classmethod
    def _can_access(cls, user, email: dict) -> bool:
        actors = set()
        actors.add(email.get('from'))
        actors.update(email.get('to', []))
        actors.update(email.get('cc', []))
        actors.update(email.get('bcc', []))

        return user.email in actors

    @classmethod
    def from_request(cls, email_store: EmailStore):
        action_name = request.args.get('action')
        form = cls._new_instance_for(action_name)
        if not form:
            return None

        form._populate(email_store)

        return form


class ToEmailForm(NewEmailForm):
    action_name = 'to'

    # noinspection PyUnusedLocal
    def _populate(self, email_store: EmailStore):
        if not self.to.data:
            self.to.data = request.args.get('to', '')


class ReplyEmailForm(NewEmailForm):
    action_name = 'reply'

    def _populate(self, email_store: EmailStore):
        email = self._get_reference_email(email_store)
        if not email:
            return

        if not self.to.data:
            self.to.data = email.get('from', '')

        if not self.subject.data:
            self.subject.data = 'Re: {}'.format(email.get('subject', ''))

        if not self.body.data:
            self.body.data = render_template('emails/reply.html', email=email)


class ReplyAllEmailForm(NewEmailForm):
    action_name = 'reply_all'

    def _populate(self, email_store: EmailStore):
        email = self._get_reference_email(email_store)
        if not email:
            return

        if not self.to.data:
            self.to.data = _join_emails(email.get('from'), *email.get('cc', []))

        if not self.subject.data:
            self.subject.data = 'Re: {}'.format(email.get('subject', ''))

        if not self.body.data:
            self.body.data = render_template('emails/reply.html', email=email)


class ForwardEmailForm(NewEmailForm):
    action_name = 'forward'

    def _populate(self, email_store: EmailStore):
        email = self._get_reference_email(email_store)
        if not email:
            return

        if not self.subject.data:
            self.subject.data = 'Fwd: {}'.format(email.get('subject', ''))

        if not self.body.data:
            self.body.data = render_template('emails/forward.html', email=email)

        self._set_forwarded_attachments(email)

    def _set_forwarded_attachments(self, email):
        attachment_filenames = [attachment.get('filename') for attachment in email.get('attachments', [])]

        self.forwarded_attachments.choices = [(filename, filename) for filename in attachment_filenames]

        self.forwarded_attachments.render_kw = {
            'size': len(attachment_filenames),
        }

        if not self.forwarded_attachments.data:
            self.forwarded_attachments.data = [filename for filename in attachment_filenames]

    def as_dict(self, email_store: EmailStore):
        form = super().as_dict(email_store)
        new_attachments = form.get('attachments', [])

        reference_email = self._get_reference_email(email_store)

        forwarded_attachments = [
            attachment for attachment in reference_email.get('attachments', [])
            if attachment.get('filename') in self.forwarded_attachments.data
        ]

        form['attachments'] = forwarded_attachments + new_attachments

        return form


def _attachments_as_dict(filestorages: Iterable[FileStorage]) \
        -> Iterable[dict]:

    for filestorage in filestorages:
        filename = filestorage.filename
        content = filestorage.stream.read()

        formatted_content = _format_attachment(filename, content)

        if filename and formatted_content:
            yield {'filename': filename, 'content': formatted_content}


def _format_attachment(filename: str, content: bytes) -> bytes:
    attachment_type = guess_type(filename)[0]

    if not attachment_type:
        return content

    if 'image' in attachment_type.lower():
        content = _change_image_size(content)

    return content


def _is_already_small(size: Tuple[int, int]) -> bool:
    width, height = size
    return width <= ImageDimensions.MAX_WIDTH_IMAGES and height <= ImageDimensions.MAX_HEIGHT_IMAGES


def _change_image_size(image_content_bytes: bytes) -> bytes:
    image_bytes = BytesIO(image_content_bytes)
    image_bytes.seek(0)
    image = Image.open(image_bytes)

    if _is_already_small(image.size):
        return image_content_bytes

    new_size = (ImageDimensions.MAX_WIDTH_IMAGES, ImageDimensions.MAX_HEIGHT_IMAGES)
    image.thumbnail(new_size, Image.ANTIALIAS)
    new_image = BytesIO()
    image.save(new_image, image.format)
    new_image.seek(0)
    new_image_bytes = new_image.read()
    return new_image_bytes


def _is_local_message(address: str) -> bool:
    host = address.split('@')[-1]
    return host.lower() == AppConfig.CLIENT_EMAIL_HOST.lower()


def _join_emails(*emails: str) -> str:
    delimiter = '{0} '.format(AppConfig.EMAIL_ADDRESS_DELIMITER)
    return delimiter.join(email for email in emails if email)


def _split_emails(emails: Optional[str]) -> List[str]:
    if not emails:
        return []

    addresses = emails.split(AppConfig.EMAIL_ADDRESS_DELIMITER)
    return [address.strip() for address in addresses]



#!/usr/bin/env python3
from glob import glob
from os import getenv
from os import remove
from os.path import join
from pathlib import Path
from threading import Event

from flask_migrate import MigrateCommand
from flask_script import Manager
from flask_security.utils import hash_password
from watchdog.events import FileSystemEvent
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from opwen_email_client.webapp import app
from opwen_email_client.webapp.actions import RestartAppComponent
from opwen_email_client.webapp.config import AppConfig

manager = Manager(app)
manager.add_command('db', MigrateCommand)


@manager.command
def devserver():
    templates_directory = join(app.root_path, app.template_folder)
    templates_glob = join(templates_directory, '**', '*.html')
    reload_server_if_changed = glob(templates_glob, recursive=True)

    port = int(getenv('WEBAPP_PORT', '5000'))
    host = getenv('HOST', '127.0.0.1')

    app.run(debug=True, extra_files=reload_server_if_changed, host=host, port=port)  # nosec


@manager.command
def resetdb():
    remove(AppConfig.LOCAL_EMAIL_STORE)
    remove(AppConfig.SQLITE_PATH)


@manager.option('-d', '--directory', required=True)
def restarter(directory):
    class Restarter(FileSystemEventHandler):
        def on_created(self, event: FileSystemEvent):
            restart = RestartAppComponent(restart_path=event.src_path)
            restart()

    Path(directory).mkdir(exist_ok=True, parents=True)
    observer = Observer()
    observer.schedule(Restarter(), directory)
    observer.start()
    try:
        Event().wait()
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


@manager.option('-n', '--name', required=True)
@manager.option('-p', '--password', required=True)
def createadmin(name, password):
    user_datastore = app.ioc.user_store
    email = '{}@{}'.format(name, AppConfig.CLIENT_EMAIL_HOST)
    password = hash_password(password)

    user = user_datastore.r.find_user(email=email)

    if user is None:
        user_datastore.w.create_user(email=email, password=password, is_admin=True)
    else:
        user.is_admin = True
        user.password = password
        user_datastore.w.put(user)

    user_datastore.w.commit()


if __name__ == '__main__':
    manager.run()



from opwen_email_server import config
from opwen_email_server.services.auth import AzureAuth
from opwen_email_server.services.auth import NoAuth
from opwen_email_server.services.storage import AzureFileStorage
from opwen_email_server.services.storage import AzureObjectsStorage
from opwen_email_server.services.storage import AzureObjectStorage
from opwen_email_server.services.storage import AzureTextStorage
from opwen_email_server.utils.collections import singleton
from opwen_email_server.utils.unique import NewGuid


def get_no_auth() -> NoAuth:
    return NoAuth()


@singleton
def get_guid_source() -> NewGuid:
    return NewGuid(config.RANDOM_SEED)


@singleton
def get_auth() -> AzureAuth:
    return AzureAuth(storage=AzureObjectStorage(
        account=config.TABLES_ACCOUNT,
        key=config.TABLES_KEY,
        host=config.TABLES_HOST,
        secure=config.TABLES_SECURE,
        container=config.CONTAINER_AUTH,
        provider=config.STORAGE_PROVIDER,
    ))


@singleton
def get_client_storage() -> AzureObjectsStorage:
    return AzureObjectsStorage(
        file_storage=AzureFileStorage(
            account=config.CLIENT_STORAGE_ACCOUNT,
            key=config.CLIENT_STORAGE_KEY,
            host=config.CLIENT_STORAGE_HOST,
            secure=config.CLIENT_STORAGE_SECURE,
            container=config.CONTAINER_CLIENT_PACKAGES,
            provider=config.STORAGE_PROVIDER,
        ),
        resource_id_source=get_guid_source(),
    )


@singleton
def get_raw_email_storage() -> AzureTextStorage:
    return AzureTextStorage(
        account=config.BLOBS_ACCOUNT,
        key=config.BLOBS_KEY,
        host=config.BLOBS_HOST,
        secure=config.BLOBS_SECURE,
        container=config.CONTAINER_SENDGRID_MIME,
        provider=config.STORAGE_PROVIDER,
    )


@singleton
def get_email_storage() -> AzureObjectStorage:
    return AzureObjectStorage(
        account=config.BLOBS_ACCOUNT,
        key=config.BLOBS_KEY,
        host=config.BLOBS_HOST,
        secure=config.BLOBS_SECURE,
        container=config.CONTAINER_EMAILS,
        provider=config.STORAGE_PROVIDER,
    )


@singleton
def get_user_storage() -> AzureObjectStorage:
    return AzureObjectStorage(
        account=config.TABLES_ACCOUNT,
        key=config.TABLES_KEY,
        host=config.TABLES_HOST,
        secure=config.TABLES_SECURE,
        container=config.CONTAINER_USERS,
        provider=config.STORAGE_PROVIDER,
        case_sensitive=False,
    )


@singleton
def get_mailbox_storage() -> AzureTextStorage:
    return AzureTextStorage(
        account=config.BLOBS_ACCOUNT,
        key=config.BLOBS_KEY,
        host=config.BLOBS_HOST,
        secure=config.BLOBS_SECURE,
        container=config.CONTAINER_MAILBOX,
        provider=config.STORAGE_PROVIDER,
        case_sensitive=False,
    )


@singleton
def get_pending_storage() -> AzureTextStorage:
    return AzureTextStorage(
        account=config.TABLES_ACCOUNT,
        key=config.TABLES_KEY,
        host=config.TABLES_HOST,
        secure=config.TABLES_SECURE,
        container=config.CONTAINER_PENDING,
        provider=config.STORAGE_PROVIDER,
    )



from environs import Env

from opwen_email_server.utils.string import urlsafe

env = Env()

RANDOM_SEED = env.int('LOKOLE_RANDOM_SEED', None)

STORAGE_PROVIDER = env('LOKOLE_STORAGE_PROVIDER', 'AZURE_BLOBS')

BLOBS_ACCOUNT = env('LOKOLE_EMAIL_SERVER_AZURE_BLOBS_NAME', '')
BLOBS_KEY = env('LOKOLE_EMAIL_SERVER_AZURE_BLOBS_KEY', '')
BLOBS_HOST = env('LOKOLE_EMAIL_SERVER_AZURE_BLOBS_HOST', '') or None
BLOBS_SECURE = env.bool('LOKOLE_EMAIL_SERVER_AZURE_BLOBS_SECURE', True)

TABLES_ACCOUNT = env('LOKOLE_EMAIL_SERVER_AZURE_TABLES_NAME', '')
TABLES_KEY = env('LOKOLE_EMAIL_SERVER_AZURE_TABLES_KEY', '')
TABLES_HOST = env('LOKOLE_EMAIL_SERVER_AZURE_TABLES_HOST', '') or None
TABLES_SECURE = env.bool('LOKOLE_EMAIL_SERVER_AZURE_TABLES_SECURE', True)

CLIENT_STORAGE_ACCOUNT = env('LOKOLE_CLIENT_AZURE_STORAGE_NAME', '')
CLIENT_STORAGE_KEY = env('LOKOLE_CLIENT_AZURE_STORAGE_KEY', '')
CLIENT_STORAGE_HOST = env('LOKOLE_CLIENT_AZURE_STORAGE_HOST', '') or None
CLIENT_STORAGE_SECURE = env.bool('LOKOLE_CLIENT_AZURE_STORAGE_SECURE', True)

resource_suffix = env('LOKOLE_RESOURCE_SUFFIX', '')

CONTAINER_CLIENT_PACKAGES = f'compressedpackages{resource_suffix}'
CONTAINER_EMAILS = f'emails{resource_suffix}'
CONTAINER_MAILBOX = f'mailbox{resource_suffix}'
CONTAINER_USERS = f'users{resource_suffix}'
CONTAINER_SENDGRID_MIME = f'sendgridinboundemails{resource_suffix}'
CONTAINER_PENDING = f'pendingemails{resource_suffix}'
CONTAINER_AUTH = f'clientsauth{resource_suffix}'

REGISTER_CLIENT_QUEUE = f'register{resource_suffix}'
INBOUND_STORE_QUEUE = f'inbound{resource_suffix}'
WRITTEN_STORE_QUEUE = f'written{resource_suffix}'
PROCESS_SERVICE_QUEUE = f'service{resource_suffix}'
SEND_QUEUE = f'send{resource_suffix}'
MAILBOX_RECEIVED_QUEUE = f'mailboxreceived{resource_suffix}'
MAILBOX_SENT_QUEUE = f'mailboxsent{resource_suffix}'

SENDGRID_MAX_RETRIES = env.int('LOKOLE_SENDGRID_MAX_RETRIES', 20)
SENDGRID_RETRY_INTERVAL_SECONDS = env.float('LOKOLE_SENDGRID_RETRY_INTERVAL_SECONDS', 5)
SENDGRID_KEY = env('LOKOLE_SENDGRID_KEY', '')
DNS_ACCOUNT = env('LOKOLE_CLOUDFLARE_USER', '')
DNS_SECRET = env('LOKOLE_CLOUDFLARE_KEY', '')
DNS_PROVIDER = env('LOKOLE_DNS_PROVIDER', 'CLOUDFLARE')

LOG_LEVEL = env('LOKOLE_LOG_LEVEL', 'INFO')

APPINSIGHTS_LOG_LEVEL = env('LOKOLE_EMAIL_SERVER_APPINSIGHTS_LOG_LEVEL', 'WARNING')
APPINSIGHTS_KEY = env('LOKOLE_EMAIL_SERVER_APPINSIGHTS_KEY', '')
APPINSIGHTS_HOST = env('LOKOLE_EMAIL_SERVER_APPINSIGHTS_HOST', '')

REGISTRATION_USERNAME = env('LOKOLE_REGISTRATION_USERNAME', '')
REGISTRATION_PASSWORD = env('LOKOLE_REGISTRATION_PASSWORD', '')
REGISTRATION_GITHUB_ORGANIZATION = env('LOKOLE_REGISTRATION_GITHUB_ORGANIZATION', 'ascoderu')
REGISTRATION_GITHUB_TEAM = env('LOKOLE_REGISTRATION_GITHUB_ORGANIZATION', 'lokole-registration')

MAX_WIDTH_IMAGES = env.int('LOKOLE_MAX_WIDTH_EMAIL_IMAGES', 200)
MAX_HEIGHT_IMAGES = env.int('LOKOLE_MAX_HEIGHT_EMAIL_IMAGES', 200)

if env('LOKOLE_QUEUE_BROKER_SCHEME', '') == 'azureservicebus':
    QUEUE_BROKER = 'azureservicebus://{username}:{password}@{host}'.format(
        username=urlsafe(env('LOKOLE_EMAIL_SERVER_QUEUES_SAS_NAME')),
        password=urlsafe(env('LOKOLE_EMAIL_SERVER_QUEUES_SAS_KEY')),
        host=urlsafe(env('LOKOLE_EMAIL_SERVER_QUEUES_NAMESPACE')))
else:
    QUEUE_BROKER = env('LOKOLE_QUEUE_BROKER_URL', '')



from typing_extensions import Final  # noqa: F401

CLIENT_DELETED = 'client_deleted'  # type: Final
CLIENT_FETCHED = 'client_fetched'  # type: Final
CLIENTS_FETCHED = 'clients_fetched'  # type: Final
CLIENT_CREATED = 'client_created'  # type: Final
NEW_CLIENT_REGISTERED = 'new_client_registered'  # type: Final
UNREGISTERED_CLIENT = 'unregistered_client'  # type: Final
UNKNOWN_USER = 'unknown_user'  # type: Final
BAD_PASSWORD = 'bad_password'  # type: Final  # nosec
MISSING_SCOPES = 'missing_scopes'  # type: Final
UNKNOWN_COMPRESSION_FORMAT = 'unknown_compression_format'  # type: Final
EMAILS_DELIVERED_TO_CLIENT = 'emails_delivered_to_client'  # type: Final
EMAILS_FORMATTED_FOR_CLIENT = 'emails_formatted_for_client'  # type: Final
EMAILS_RECEIVED_FROM_CLIENT = 'emails_received_from_client'  # type: Final
EMAIL_RECEIVED_FOR_CLIENT = 'email_received_for_client'  # type: Final
EMAIL_DELIVERED_FROM_CLIENT = 'email_delivered_from_client'  # type: Final
EMAIL_STORED_FOR_CLIENT = 'email_stored_for_client'  # type: Final
EMAIL_STORED_FROM_CLIENT = 'email_stored_from_client'  # type: Final
USER_STORED_FROM_CLIENT = 'user_stored_from_client'  # type: Final
MAILBOX_EMAIL_INDEXED = 'mailbox_email_indexed'  # type: Final



from pathlib import Path
from subprocess import Popen  # nosec
from time import sleep
from typing import IO


def _dialer_is_connected(log_path: str) -> bool:
    with open(log_path, 'rb') as fobj:
        for line in fobj:
            if line.startswith(b'--> secondary DNS address'):
                return True
    return False


def _start_dialer(config: Path, log_file: IO) -> Popen:
    return Popen(['/usr/bin/wvdial', '--config', str(config.absolute())], stderr=log_file)


def dialup(config: Path, log: Path, max_retries: int, poll_seconds: int) -> Popen:
    with log.open(mode='w+b') as log:
        connection = _start_dialer(config, log)

        while not _dialer_is_connected(log.name):
            if connection.poll() is not None:
                connection.terminate()
                raise ValueError('Invalid wvdial configuration')

            if max_retries <= 0:
                connection.terminate()
                raise ValueError('Modem taking too long to connect')

            sleep(poll_seconds)
            max_retries -= 1

        return connection



from opwen_email_server.utils.log import LogMixin

ECHO_ADDRESS = 'echo@bot.lokole.ca'


class EchoEmailFormatter(LogMixin):
    def __call__(self, email: dict) -> dict:
        email['to'] = [email['from']]
        email['from'] = ECHO_ADDRESS
        return email



from typing import List

from flask import Flask
from flask_migrate import Migrate
from flask_security import RoleMixin
from flask_security import SQLAlchemyUserDatastore
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import and_
from sqlalchemy.exc import OperationalError

from opwen_email_client.domain.email.user_store import User
from opwen_email_client.domain.email.user_store import UserStore
from opwen_email_client.webapp.config import AppConfig

_db = SQLAlchemy()

# noinspection PyUnresolvedReferences
_roles_users = _db.Table(
    'roles_users',
    _db.Column('user_id', _db.Integer(), _db.ForeignKey('user.id')),
    _db.Column('role_id', _db.Integer(), _db.ForeignKey('role.id')),
)


# noinspection PyUnresolvedReferences
class _User(_db.Model, User):
    __tablename__ = 'user'

    id = _db.Column(_db.Integer(), primary_key=True)
    email = _db.Column(_db.String(255), unique=True, index=True)
    password = _db.Column(_db.String(255), nullable=False)
    active = _db.Column(_db.Boolean(), default=True)
    last_login_at = _db.Column(_db.DateTime())
    current_login_at = _db.Column(_db.DateTime())
    last_login_ip = _db.Column(_db.String(128))
    current_login_ip = _db.Column(_db.String(128))
    login_count = _db.Column(_db.Integer())
    timezone_offset_minutes = _db.Column(_db.Integer(), nullable=False, default=0)
    language = _db.Column(_db.String(8))
    roles = _db.relationship('_Role', secondary=_roles_users, backref=_db.backref('users', lazy='dynamic'))
    synced = _db.Column(_db.Boolean(), default=False)
    is_admin = _db.Column(_db.Boolean(), default=False)


# noinspection PyUnresolvedReferences
class _Role(_db.Model, RoleMixin):
    __tablename__ = 'role'

    id = _db.Column(_db.Integer(), primary_key=True)
    name = _db.Column(_db.String(32), unique=True)
    description = _db.Column(_db.String(255))


_migrate = Migrate()


class FlaskLoginUserStore(UserStore):
    def __init__(self):
        store = SQLAlchemyUserDatastore(_db, _User, _Role)
        super().__init__(read=store, write=store)
        self._app = None

    def init_app(self, app: Flask):
        with app.app_context():
            _db.init_app(app)
            _migrate.init_app(app, _db)

            try:
                _db.create_all()
            except OperationalError:
                pass

        self._app = app

    def fetch_all(self, user: User) -> List[User]:
        with self._app.app_context():
            return _User.query.all()

    def fetch_pending(self) -> List[User]:
        is_not_synced = _User.synced == False  # noqa: E712
        is_non_admin = _User.email != AppConfig.ADMIN_INBOX

        with self._app.app_context():
            return _User.query\
                .filter(and_(is_not_synced, is_non_admin))\
                .all()



from cached_property import cached_property
from libcloud.common.types import LibcloudError
from libcloud.dns.base import DNSDriver
from libcloud.dns.base import Zone
from libcloud.dns.providers import get_driver
from libcloud.dns.types import Provider
from libcloud.dns.types import RecordType

from opwen_email_server.constants.sendgrid import MX_RECORD
from opwen_email_server.utils.log import LogMixin


class _MxRecords(LogMixin):
    def __init__(self, account: str, secret: str, provider: str) -> None:
        self._account = account
        self._secret = secret
        self._provider = getattr(Provider, provider)

    @cached_property
    def _driver(self) -> DNSDriver:
        driver = get_driver(self._provider)
        return driver(self._account, self._secret)

    def __call__(self, domain: str) -> None:
        if not self._secret:
            self.log_warning('No key, skipping MX setup for %s', domain)
            return

        domain_parts = domain.split('.')
        client_name = domain_parts[0]
        zone_name = '.'.join(domain_parts[1:])

        zone = next(zone for zone in self._driver.iterate_zones() if zone.domain == zone_name)

        self._run(client_name, zone)

    def _run(self, client_name: str, zone: Zone) -> None:
        raise NotImplementedError  # pragma: no cover


class DeleteMxRecords(_MxRecords):
    def _run(self, client_name: str, zone: Zone) -> None:
        try:
            record = next(record for record in self._driver.iterate_records(zone) if record.name == client_name)
        except StopIteration:
            self.log_warning('No MX records for client %s.%s exist', client_name, zone.domain)
        else:
            self._driver.delete_record(record)
            self.log_info('Deleted MX records for client %s.%s', client_name, zone.domain)


class SetupMxRecords(_MxRecords):
    def _run(self, client_name: str, zone: Zone) -> None:
        try:
            self._driver.create_record(
                zone=zone,
                name=client_name,
                type=RecordType.MX,
                data=MX_RECORD,
                extra={'priority': 0},
            )
        except LibcloudError as ex:
            if 'record already exists' not in ex.value:
                raise ex
            self.log_warning('MX records for client %s.%s already exist', client_name, zone.domain)
        else:
            self.log_info('Set up MX records for client %s.%s', client_name, zone.domain)



from celery import Celery

from opwen_email_server import config
from opwen_email_server.actions import IndexReceivedEmailForMailbox
from opwen_email_server.actions import IndexSentEmailForMailbox
from opwen_email_server.actions import ProcessServiceEmail
from opwen_email_server.actions import RegisterClient
from opwen_email_server.actions import SendOutboundEmails
from opwen_email_server.actions import StoreInboundEmails
from opwen_email_server.actions import StoreWrittenClientEmails
from opwen_email_server.integration.azure import get_auth
from opwen_email_server.integration.azure import get_client_storage
from opwen_email_server.integration.azure import get_email_storage
from opwen_email_server.integration.azure import get_guid_source
from opwen_email_server.integration.azure import get_mailbox_storage
from opwen_email_server.integration.azure import get_pending_storage
from opwen_email_server.integration.azure import get_raw_email_storage
from opwen_email_server.integration.azure import get_user_storage
from opwen_email_server.mailers import REGISTRY
from opwen_email_server.services.dns import SetupMxRecords
from opwen_email_server.services.sendgrid import SendSendgridEmail
from opwen_email_server.services.sendgrid import SetupSendgridMailbox

celery = Celery(broker=config.QUEUE_BROKER)


@celery.task(ignore_result=True)
def register_client(domain: str, owner: str) -> None:
    action = RegisterClient(
        auth=get_auth(),
        client_storage=get_client_storage(),
        setup_mailbox=SetupSendgridMailbox(
            key=config.SENDGRID_KEY,
            max_retries=config.SENDGRID_MAX_RETRIES,
            retry_interval_seconds=config.SENDGRID_RETRY_INTERVAL_SECONDS,
        ),
        setup_mx_records=SetupMxRecords(
            account=config.DNS_ACCOUNT,
            secret=config.DNS_SECRET,
            provider=config.DNS_PROVIDER,
        ),
        client_id_source=get_guid_source(),
    )

    action(domain, owner)


@celery.task(ignore_result=True)
def index_received_email_for_mailbox(resource_id: str) -> None:
    action = IndexReceivedEmailForMailbox(
        email_storage=get_email_storage(),
        mailbox_storage=get_mailbox_storage(),
    )

    action(resource_id)


@celery.task(ignore_result=True)
def index_sent_email_for_mailbox(resource_id: str) -> None:
    action = IndexSentEmailForMailbox(
        email_storage=get_email_storage(),
        mailbox_storage=get_mailbox_storage(),
    )

    action(resource_id)


@celery.task(ignore_result=True)
def inbound_store(resource_id: str) -> None:
    action = StoreInboundEmails(
        raw_email_storage=get_raw_email_storage(),
        email_storage=get_email_storage(),
        pending_storage=get_pending_storage(),
        next_task=index_received_email_for_mailbox.delay,
    )

    action(resource_id)


def send_and_index_email(resource_id: str) -> None:
    send.delay(resource_id)
    index_sent_email_for_mailbox.delay(resource_id)
    index_received_email_for_mailbox.delay(resource_id)


@celery.task(ignore_result=True)
def written_store(resource_id: str) -> None:
    action = StoreWrittenClientEmails(
        client_storage=get_client_storage(),
        email_storage=get_email_storage(),
        user_storage=get_user_storage(),
        next_task=send_and_index_email,
    )

    action(resource_id)


@celery.task(ignore_result=True)
def send(resource_id: str) -> None:
    action = SendOutboundEmails(
        email_storage=get_email_storage(),
        send_email=SendSendgridEmail(key=config.SENDGRID_KEY),
    )

    action(resource_id)


@celery.task(ignore_result=True)
def process_service_email(resource_id: str) -> None:
    action = ProcessServiceEmail(
        raw_email_storage=get_raw_email_storage(),
        email_storage=get_email_storage(),
        registry=REGISTRY,
        next_task=send_and_index_email,
    )

    action(resource_id)


def _fqn(task):
    return f'{__name__}.{task.__name__}'


task_routes = {
    _fqn(register_client): {'queue': config.REGISTER_CLIENT_QUEUE},
    _fqn(index_received_email_for_mailbox): {'queue': config.MAILBOX_RECEIVED_QUEUE},
    _fqn(index_sent_email_for_mailbox): {'queue': config.MAILBOX_SENT_QUEUE},
    _fqn(process_service_email): {'queue': config.PROCESS_SERVICE_QUEUE},
    _fqn(inbound_store): {'queue': config.INBOUND_STORE_QUEUE},
    _fqn(written_store): {'queue': config.WRITTEN_STORE_QUEUE},
    _fqn(send): {'queue': config.SEND_QUEUE}
}

celery.conf.update(task_routes=task_routes)

if __name__ == '__main__':
    celery.start()



from copy import deepcopy
from datetime import datetime
from datetime import timezone
from email.utils import mktime_tz
from email.utils import parsedate_tz
from io import BytesIO
from itertools import chain
from mimetypes import guess_type
from typing import Callable
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from bs4 import BeautifulSoup
from PIL import Image
from pyzmail import PyzMessage
from pyzmail.parse import MailPart
from requests import Response
from requests import get as http_get

from opwen_email_server.config import MAX_HEIGHT_IMAGES
from opwen_email_server.config import MAX_WIDTH_IMAGES
from opwen_email_server.utils.log import LogMixin
from opwen_email_server.utils.serialization import to_base64


def _parse_body(message: PyzMessage, default_charset: str = 'ascii') -> str:
    body_parts = (message.html_part, message.text_part)
    for part in body_parts:
        if part is None:
            continue
        payload = part.get_payload()
        if payload is None:
            continue
        charset = part.charset or default_charset
        return payload.decode(charset, errors='replace')
    return ''


def _parse_attachments(mailparts: Iterable[MailPart]) -> Iterable[dict]:
    attachment_parts = (part for part in mailparts if not part.is_body)
    for part in attachment_parts:
        filename = part.sanitized_filename
        payload = part.get_payload()
        attachment_id = part.content_id
        if filename and payload:
            attachment = {'filename': filename, 'content': payload}
            if attachment_id:
                attachment['cid'] = attachment_id
            yield attachment


def _parse_addresses(message: PyzMessage, address_type: str) -> List[str]:
    return sorted(email for _, email in message.get_addresses(address_type) if email)


def _parse_address(message: PyzMessage, address_type: str) -> Optional[str]:
    return next(iter(_parse_addresses(message, address_type)), None)


def _parse_sent_at(message: PyzMessage) -> Optional[str]:
    rfc_822 = message.get_decoded_header('date')
    if not rfc_822:
        return None
    date_tz = parsedate_tz(rfc_822)
    if not date_tz:
        return None
    timestamp = mktime_tz(date_tz)
    # noinspection PyUnresolvedReferences
    date_utc = datetime.fromtimestamp(timestamp, timezone.utc)
    return date_utc.strftime('%Y-%m-%d %H:%M')


def parse_mime_email(mime_email: str) -> dict:
    message = PyzMessage.factory(mime_email)

    return {
        'sent_at': _parse_sent_at(message),
        'to': _parse_addresses(message, 'to'),
        'cc': _parse_addresses(message, 'cc'),
        'bcc': _parse_addresses(message, 'bcc'),
        'from': _parse_address(message, 'from'),
        'subject': message.get_subject(),
        'body': _parse_body(message),
        'attachments': list(_parse_attachments(message.mailparts)),
    }


def format_attachments(email: dict) -> dict:
    attachments = email.get('attachments', [])

    if not attachments:
        return email

    formatted_attachments = deepcopy(attachments)
    is_any_attachment_changed = False

    for i, attachment in enumerate(attachments):
        filename = attachment.get('filename', '')
        content = attachment.get('content', b'')
        formatted_content = _format_attachment(filename, content)

        if content != formatted_content:
            formatted_attachments[i]['content'] = formatted_content
            is_any_attachment_changed = True

    if not is_any_attachment_changed:
        return email

    new_email = dict(email)
    new_email['attachments'] = formatted_attachments
    return new_email


def _format_attachment(filename: str, content: bytes) -> bytes:
    attachment_type = guess_type(filename)[0]

    if not attachment_type:
        return content

    if 'image' in attachment_type.lower():
        content = _change_image_size(content)

    return content


def get_recipients(email: dict) -> Iterable[str]:
    return chain(email.get('to') or [], email.get('cc') or [], email.get('bcc') or [])


def get_domains(email: dict) -> Iterable[str]:
    return frozenset(get_domain(address) for address in get_recipients(email))


def get_domain(address: str) -> str:
    return address.split('@')[-1]


def ensure_has_sent_at(email: dict):
    if not email.get('sent_at'):
        email['sent_at'] = datetime.utcnow().strftime('%Y-%m-%d %H:%M')


def _get_image_type(response: Response, url: str) -> Optional[str]:
    content_type = response.headers.get('Content-Type')
    if not content_type:
        content_type = guess_type(url)[0]
    return content_type


def _is_already_small(size: Tuple[int, int]) -> bool:
    width, height = size
    return width <= MAX_WIDTH_IMAGES and height <= MAX_HEIGHT_IMAGES


def _change_image_size(image_content_bytes: bytes) -> bytes:
    image_bytes = BytesIO(image_content_bytes)
    image_bytes.seek(0)
    image = Image.open(image_bytes)

    if _is_already_small(image.size):
        return image_content_bytes

    new_size = (MAX_WIDTH_IMAGES, MAX_HEIGHT_IMAGES)
    image.thumbnail(new_size, Image.ANTIALIAS)
    new_image = BytesIO()
    image.save(new_image, image.format)
    new_image.seek(0)
    new_image_bytes = new_image.read()
    return new_image_bytes


def _fetch_image_to_base64(image_url: str) -> Optional[str]:
    response = http_get(image_url)
    if not response.ok:
        return None

    image_type = _get_image_type(response, image_url)
    if not image_type:
        return None

    if not response.content:
        return None

    small_image_bytes = _change_image_size(response.content)
    small_image_base64 = to_base64(small_image_bytes)
    return f'data:{image_type};base64,{small_image_base64}'


def _is_valid_url(url: Optional[str]) -> bool:
    if not url:
        return False
    has_http_prefix = url.startswith('http://')
    has_https_prefix = url.startswith('https://')
    return has_http_prefix or has_https_prefix


def format_inline_images(email: dict, on_error: Callable) -> dict:
    email_body = email.get('body', '')
    if not email_body:
        return email

    soup = BeautifulSoup(email_body, 'html.parser')
    image_tags = soup.find_all('img')
    if not image_tags:
        return email

    for image_tag in image_tags:
        image_url = image_tag.get('src')

        if not _is_valid_url(image_url):
            continue

        try:
            encoded_image = _fetch_image_to_base64(image_url)
        except Exception as ex:
            on_error('Unable to inline image %s: %s', image_url, ex)
        else:
            if encoded_image:
                image_tag['src'] = encoded_image

    new_email = dict(email)
    new_email['body'] = str(soup)
    return new_email


class MimeEmailParser(LogMixin):
    def __call__(self, mime_email: str) -> dict:
        email = parse_mime_email(mime_email)
        email = format_attachments(email)
        email = format_inline_images(email, self.log_warning)
        return email



#!/usr/bin/env python3

from applicationinsights.flask.ext import AppInsights
from connexion import App
from flask import Flask
from flask_cors import CORS

from opwen_email_server import config

_hosts = ['127.0.0.1', '0.0.0.0']  # nosec

_host = _hosts[0]
_port = 8080
_ui = False


def build_app(apis, host=_host, port=_port, ui=_ui):
    app = App(__name__, host=host, port=port, server='flask', options={'swagger_ui': ui})

    for api in apis:
        app.add_api(api)

    _configure_flask(app.app)

    return app


def _configure_flask(app: Flask):
    app.config['APPINSIGHTS_INSTRUMENTATIONKEY'] = config.APPINSIGHTS_KEY
    app.config['APPINSIGHTS_ENDPOINT_URI'] = config.APPINSIGHTS_HOST
    app.config['APPINSIGHTS_DISABLE_TRACE_LOGGING'] = True
    app.config['APPINSIGHTS_DISABLE_REQUEST_LOGGING'] = True
    AppInsights(app)

    CORS(app)


def _cli():
    from argparse import ArgumentParser
    from argparse import FileType

    parser = ArgumentParser()
    parser.add_argument('--host', choices=_hosts, default=_host)
    parser.add_argument('--port', type=int, default=_port)
    parser.add_argument('--ui', action='store_true', default=_ui)
    parser.add_argument('apis', nargs='+', type=FileType('r'))
    args = parser.parse_args()

    apis = []
    for fobj in args.apis:
        apis.append(fobj.name)
        fobj.close()

    app = build_app(apis, args.host, args.port, args.ui)

    app.run()


if __name__ == '__main__':
    _cli()



from opwen_email_server import config
from opwen_email_server.actions import CalculateNumberOfUsersMetric
from opwen_email_server.actions import CalculatePendingEmailsMetric
from opwen_email_server.actions import CreateClient
from opwen_email_server.actions import DeleteClient
from opwen_email_server.actions import DownloadClientEmails
from opwen_email_server.actions import GetClient
from opwen_email_server.actions import ListClients
from opwen_email_server.actions import Ping
from opwen_email_server.actions import ReceiveInboundEmail
from opwen_email_server.actions import UploadClientEmails
from opwen_email_server.integration.azure import get_auth
from opwen_email_server.integration.azure import get_client_storage
from opwen_email_server.integration.azure import get_email_storage
from opwen_email_server.integration.azure import get_mailbox_storage
from opwen_email_server.integration.azure import get_no_auth
from opwen_email_server.integration.azure import get_pending_storage
from opwen_email_server.integration.azure import get_raw_email_storage
from opwen_email_server.integration.azure import get_user_storage
from opwen_email_server.integration.celery import inbound_store
from opwen_email_server.integration.celery import process_service_email
from opwen_email_server.integration.celery import register_client
from opwen_email_server.integration.celery import written_store
from opwen_email_server.services.auth import AnyOfBasicAuth
from opwen_email_server.services.auth import BasicAuth
from opwen_email_server.services.auth import GithubBasicAuth
from opwen_email_server.services.dns import DeleteMxRecords
from opwen_email_server.services.sendgrid import DeleteSendgridMailbox

email_receive = ReceiveInboundEmail(
    auth=get_auth(),
    raw_email_storage=get_raw_email_storage(),
    next_task=inbound_store.delay,
)

receive_service_email = ReceiveInboundEmail(
    auth=get_no_auth(),
    raw_email_storage=get_raw_email_storage(),
    next_task=process_service_email,
)

client_write = UploadClientEmails(
    auth=get_auth(),
    next_task=written_store.delay,
)

client_read = DownloadClientEmails(
    auth=get_auth(),
    client_storage=get_client_storage(),
    email_storage=get_email_storage(),
    pending_storage=get_pending_storage(),
)

client_create = CreateClient(
    auth=get_auth(),
    task=register_client.delay,
)

client_list = ListClients(auth=get_auth())

client_get = GetClient(
    auth=get_auth(),
    client_storage=get_client_storage(),
)

client_delete = DeleteClient(
    auth=get_auth(),
    delete_mailbox=DeleteSendgridMailbox(key=config.SENDGRID_KEY),
    delete_mx_records=DeleteMxRecords(
        account=config.DNS_ACCOUNT,
        secret=config.DNS_SECRET,
        provider=config.DNS_PROVIDER,
    ),
    mailbox_storage=get_mailbox_storage(),
    pending_storage=get_pending_storage(),
    user_storage=get_user_storage(),
)

metrics_users = CalculateNumberOfUsersMetric(
    auth=get_auth(),
    user_storage=get_user_storage(),
)

metrics_pending = CalculatePendingEmailsMetric(
    auth=get_auth(),
    pending_storage=get_pending_storage(),
)

basic_auth = AnyOfBasicAuth(auths=[
    BasicAuth(users={
        config.REGISTRATION_USERNAME: {'password': config.REGISTRATION_PASSWORD},
    }),
    GithubBasicAuth(
        organization=config.REGISTRATION_GITHUB_ORGANIZATION,
        team=config.REGISTRATION_GITHUB_TEAM,
    ),
])

healthcheck = Ping()



import re
from pathlib import Path

from setuptools import find_packages
from setuptools import setup

client_package = 'opwen_email_client'
server_package = 'opwen_email_server'
here = Path(__file__).parent


def requirements_for(name):
    with (here / name).open(encoding='utf-8') as req:
        return [line.strip() for line in req]


with (here / 'README.rst').open(encoding='utf-8') as fobj:
    long_description = fobj.read()

with (here / client_package / '__init__.py').open(encoding='utf-8') as fobj:
    version = re.search(r"^__version__ = '([^']*)'",
                        fobj.read(), re.MULTILINE).group(1)

scripts = ['manage.py']

setup(
    name=client_package,
    version=version,
    author='Ascoderu',
    author_email='ascoderu.opwen@gmail.com',
    packages=find_packages(exclude=['tests*']),
    url='https://github.com/ascoderu/lokole',
    license='Apache Software License',
    description='Email client for the Lokole project: https://ascoderu.ca',
    long_description=long_description,
    include_package_data=True,
    scripts=scripts,
    install_requires=requirements_for('requirements-webapp.txt'),
    extras_require={
        server_package: requirements_for('requirements.txt'),
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: Email',
    ])
